#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/times.h>

#define MAX_SIGS    48

struct PTalarms_struct
{
	clock_t   alarm_time;
	pthread_t thread_id;
	int       signo;
};

#define _PTS  (sizeof(struct PTalarms_struct))

struct PTalarms_struct PTalarms[MAX_SIGS];
int tail;  /* tail points to next free entry */
int alarm_set = 0;

int compare(void *a, void *b);
int  PTinit(void);

void sig_alarm(int signo)
{
	int i, x;
	clock_t start_time;
	
	start_time = times(NULL);
	
	signal(SIGALRM, sig_alarm);
	
	for (i = 0; i < tail; i++) {
		x = ( PTalarms[i].alarm_time - start_time) / CLK_TCK;
		
		if (x < 1) {
			/* Send signal to thread */
			kill(PTalarms[i].thread_id, PTalarms[i].signo);
			PTalarms[i].alarm_time = -1;
			
			continue;
		}
		break;
	}
	
	qsort(PTalarms, tail, _PTS, compare);	
	
	if (PTalarms[0].alarm_time != -1) {
		alarm((PTalarms[0].alarm_time - start_time) / CLK_TCK);
		alarm_set = 1;
		
		x = (PTalarms[0].alarm_time - start_time) / CLK_TCK;
	}
	else
	  alarm_set = 0;
}

int PTinit(void)
{
	int i;
	
	signal(SIGALRM, sig_alarm);
	
	for (i = 0; i < MAX_SIGS; i++) {
		PTalarms[i].thread_id = 0;
	}

	tail = 0;
	
	setvbuf(stdin, NULL, _IOLBF, 4096 * 4);
	
	return 0;
}

int compare(void *a, void *b)
{
	/* First, -2 is always the "highest" number, this is an
	 * unset signal */
	if ((*((int*)a)) == -2)
		 return 1;
	if ((*((int*)b)) == -2)
		 return -2;
	
	/* -1 is the next highest number, this is a registered signal
	 * that has expired */
	if ((*((int*)a)) == -1)
		 return 1;
	if ((*((int*)b)) == -1)
		 return -1;
		 
	if ((*((int*)a)) <  (*((int*)b)))
	  return -1;
	if ((*((int*)a)) == (*((int*)b)))
	  return 0;
	
	return 1;
}

int main(void)
{
	char buffer[80];
	int id, signo, match, i;
	unsigned seconds;
	clock_t start_time;
	
	PTinit();
	
	while( fgets(buffer, sizeof(buffer), stdin) ) {
		if (strstr(buffer, "PTsignal")) {
			match = 0;
			sscanf(buffer, "%*s %d %d", &id, &signo);
			
			/* Is this signal already registered? */
			for (i = 0; i < tail; i++) {
				if ( PTalarms[i].thread_id == id) {
					match = 1;
					break;
				}
			}
			if (match == 0) {
				PTalarms[tail].alarm_time = -2; /* -2 is it isn't set */
				PTalarms[tail].thread_id = id;
				PTalarms[tail].signo = signo;
			}
			else {
				PTalarms[i].signo = signo;
			}
				
			tail++;
		}
			 
		if (strstr(buffer, "PTalarm")) {
			match = 0;
			sscanf(buffer, "%*s %d %u", &id, &seconds);
	
			start_time = times(NULL);  

			/* Find entry for this thread */
			for (i = 0; i < tail; i++) {
				if ( PTalarms[i].thread_id == id) {
					match = 1;
					break;
				}
			}
			if (match == 0) {
				fprintf(stderr, "PTserver - alarm being set when no signal handler registered\n");
				continue;
			}
	
			start_time = start_time + (seconds * CLK_TCK);
			
			PTalarms[i].alarm_time = start_time;
			
			qsort(PTalarms, tail, _PTS, compare);
			
			start_time = times(NULL);
			
			if (alarm_set == 0) {
				i = (PTalarms->alarm_time - start_time) / CLK_TCK;
				/*printf("i = %d\n", i);		*/
				
				alarm(i);
				
				if (i < 1)
				  alarm(1);
				
				alarm_set = 1;
			}
		}
	}
	
	fprintf(stderr, "PTserver died=======================\n");
	
	return -1;
}
